/*
 * kernel_drivers/adi/adi_net.c
 *
 * Copyright (c) 2003-2004 Lukasz Dembinski <dembol@nasa.com.pl>
 * All Rights Reserved
 * 
 * Date:	2004/01
 * Author: 	Lukasz Dembinski
 * Info:	adi_net.c core file
 * Contact:	mailto: <dembol@nasa.com.pl>
 *
 */

#include <agnix/init.h>
#include <agnix/adi/adi.h>
#include <agnix/net/net.h>
#include <agnix/irq.h>
#include <agnix/memory.h>
#include <agnix/strings.h>
#include <agnix/console.h>
#include <agnix/list.h>
#include <agnix/spinlock.h>

#define MOD_NAME		"NET: "

struct list_head 	adi_netdev_list;
spinlock_t 		adi_netdev_list_lock;

int net_devices_last_id = 0;

int adi_netdev_set_name(struct adi_netdev_s *netdev, char *if_name)
{
    int name_len;
    
    name_len = strlen(if_name);

    if (name_len < NETDEV_NAME_MAX_LEN)
	memcpy(netdev->if_name, if_name, name_len);
    else
	memcpy(netdev->if_name, if_name, NETDEV_NAME_MAX_LEN);

    return 0;
}

int adi_netdev_compose_name(struct adi_netdev_s *netdev)
{
    switch(netdev->if_layer_2.if_type) {
	case IF_TYPE_ETH:
	    return eth_device_compose_name(netdev);
    }

    return 0;
}

int adi_netdev_set_id(struct adi_netdev_s *netdev)
{
    netdev->if_id = net_devices_last_id;
    
    net_devices_last_id++;
    
    return 0;
}

struct adi_netdev_s *adi_netdev_find_by_name(char *name)
{
    struct adi_netdev_s *netdev;
    struct list_head *tmp;
    int found;
    u32 flags;

    spin_lock_irqsave(&adi_netdev_list_lock, flags);    
    
    found = 0;
    list_for_each(tmp, &adi_netdev_list) {
	netdev = list_entry(tmp, struct adi_netdev_s, list);
	
	if (!word_cmp(netdev->if_name, name, NETDEV_NAME_MAX_LEN)) {
	    found = 1;
	    break;
	}
    }
    
    spin_unlock_irqrestore(&adi_netdev_list_lock, flags);
    
    if (found)
	return netdev;

    return NULL;
}

int adi_netdev_list_add(struct adi_netdev_s *netdev)
{
    u32 flags;
    
    spin_lock_irqsave(&adi_netdev_list_lock, flags);
    list_add(&netdev->list, &adi_netdev_list);
    spin_unlock_irqrestore(&adi_netdev_list_lock, flags);
    
    return 0;
}

int adi_netdev_init(struct adi_netdev_s *netdev)
{
    adi_netdev_compose_name(netdev);

    switch(netdev->if_layer_2.if_type) {
	case IF_TYPE_ETH:
	    return eth_device_init(netdev);

	case IF_TYPE_FDDI:
	    break;

	case IF_TYPE_TR:
	    break;
    }

    return 0;
}

int adi_register_netdev(struct adi_netdev_s *netdev)
{
    adi_netdev_set_id  (netdev);
    adi_netdev_list_add(netdev);

    switch (netdev->if_layer_2.if_type) {
	case IF_TYPE_ETH:
	    return register_eth_device(netdev);
    }

    return 0;
}

struct adi_netdev_s *adi_netdev_alloc(void)
{
    return (struct adi_netdev_s *)get_free_pages(0);
}

void adi_netdev_free(struct adi_netdev_s *netdev )
{
    put_free_pages((u32)netdev, 0);
}

int adi_netdev_data_init(void)
{	
    INIT_LIST_HEAD(&adi_netdev_list);
    spin_lock_init(&adi_netdev_list_lock);

    return 0;
}

int __init adi_netdrv_init(void)
{
    printk(MOD_NAME "Initializing network drivers\n");

    adi_netdev_data_init();

#if CONFIG_DRIVERS_NET_RTL_8029
    adi_netdrv_rtl_8029_init();
#endif

    return 0;
}

void adi_netdev_print_interfaces(void)
{
    struct adi_netdev_s *netdev;
    struct list_head *tmp;
    const char *link_encaps;
    u32 flags;

    spin_lock_irqsave(&adi_netdev_list_lock, flags);    
    
    list_for_each(tmp, &adi_netdev_list) {
	netdev = list_entry(tmp, struct adi_netdev_s, list);
	
	if (netdev->if_layer_2.if_type == IF_TYPE_ETH)
	    link_encaps = "Ethernet";
	else
	if (netdev->if_layer_2.if_type == IF_TYPE_TR)
	    link_encaps = "Tokenring";
	else
	if (netdev->if_layer_2.if_type == IF_TYPE_FDDI)
	    link_encaps = "FDDI";
	else
	    link_encaps = "Unknown";
	
	printf("\n%s\t Link encapsulation:%s     HWaddr %02x:%02x:%02x:%02x:%02x:%02x\n",
	netdev->if_name, link_encaps, 
	netdev->if_layer_2.if_addr[0], netdev->if_layer_2.if_addr[1], netdev->if_layer_2.if_addr[2], 
	netdev->if_layer_2.if_addr[3], netdev->if_layer_2.if_addr[4], netdev->if_layer_2.if_addr[5]);
	
	printf("\t Inet address: %d.%d.%d.%d     MTU: %d\n", 
	netdev->if_layer_3.if_addr.address[0], netdev->if_layer_3.if_addr.address[1],
	netdev->if_layer_3.if_addr.address[2], netdev->if_layer_3.if_addr.address[3],
	netdev->if_layer_2.if_mtu); 
    }
    
    spin_unlock_irqrestore(&adi_netdev_list_lock, flags);
}
